﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ComponentModel;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.Windows.Interop;
using System.Xml;
using Buzz.MachineInterface;
using BuzzGUI.Interfaces;
using BuzzGUI.Common;

namespace videosync
{
    /// <summary>
    /// Interaction logic for SyncGUI.xaml
    /// uhm actually i ended up cramming the whole machine in there
    /// </summary>
    public partial class SyncGUI : UserControl, IMachineGUI
    {
        IMachine mac;
        SyncMachine syncmac;
        VideoWin twin;

        IntPtr instance, media, player; // libvlc struct pointers
        Vlc.Callback vlcPlayingHandler, vlcEndHandler, vlcErrorHandler;
        bool libvlcIsOpen = false;
        public bool vidIsOpen = false;
        bool firstPlay = true; // on first successful play event, pause immediadely
        bool restartOnSkip = false; // if the end of the video was passed, restart it

        public IMachine Machine
        {
            get { return mac; }
            set // machine got assigned to GUI
            {
                if (mac != null)
                {
                    // release old machine (machine deleted)
                    syncmac.Gui = null;
                    if (vidIsOpen)
                        twin.Close();
                    tempoDataGrid.ItemsSource = null;
                    syncmac = null;
                }
                mac = value;
                if (mac != null)
                {
                    // bind with new machine
                    syncmac = (SyncMachine)mac.ManagedMachine;
                    tempoDataGrid.ItemsSource = syncmac.TempoChanges;
                    tempoDataGrid.IsEnabled = syncmac.ManageTempos;
                    manageBpmCheckBox.IsChecked = syncmac.ManageTempos;
                    videoPathTextBox.Text = syncmac.VideoPath;
                    vlcDirTextBox.Text = syncmac.VLCPath;
                    volSlider.Value = syncmac.VideoVolume;
                    volLabel.Content = String.Format("{0}%", syncmac.VideoVolume);
                    vlcOutCombo.SelectedIndex = syncmac.VLCVout;
                    try
                    {
                        Vlc.libvlc_get_version();
                        vlcDirTextBox.IsEnabled = false;
                        browseVlcButton.IsEnabled = false;
                    }
                    catch (DllNotFoundException e)  { }
                    syncmac.Gui = this;
                }
            }
        }

        // constructor (not yet bound to a machine)
        public SyncGUI()
        {
            InitializeComponent();

            vlcPlayingHandler = (IntPtr vevent, IntPtr data) =>
            {
                if (!firstPlay)
                    return;
                firstPlay = false;
                if (!Global.Buzz.Playing)
                    Vlc.libvlc_media_player_set_pause(player, 1);
            };

            vlcEndHandler = (IntPtr vevent, IntPtr data) => { restartOnSkip = true; };
            vlcErrorHandler = (IntPtr vevent, IntPtr data) => 
            {
                MessageBox.Show("Unknown VLC error.", "Videosync error");
                //ReleaseVideo(); let's not crash things any further 
            };
        }

        internal void Play()
        {
            Vlc.libvlc_media_player_set_pause(player, 0);
        }

        internal void Pause()
        {
            Vlc.libvlc_media_player_set_pause(player, 1);
        }

        internal void Skip(int positionMs)
        {
            // if past the end, the video must be reloaded
            if (restartOnSkip)
            {
                restartOnSkip = false;
                firstPlay = true;
                Vlc.libvlc_media_player_set_media(player, media);
                Vlc.libvlc_media_player_play(player);
            }
            Vlc.libvlc_media_player_set_time(player, positionMs);
        }

        internal void SetVolume(int percent)
        {
            Vlc.libvlc_audio_set_volume(player, percent);
        }

        public bool TryVlcOpen()
        {
            try
            {
                Vlc.libvlc_get_version();
                return true;
            }
            catch (DllNotFoundException e)
            {
                if (System.IO.File.Exists(syncmac.VLCPath + @"\libvlc.dll"))
                    Environment.SetEnvironmentVariable("PATH", Environment.GetEnvironmentVariable("PATH") + ";" + syncmac.VLCPath);
                else
                    return false;
            }
            try
            {
                Vlc.libvlc_get_version();
                return true;
            }
            catch (DllNotFoundException e)
            {
                return false;
            }
        }

        public void OpenVideo()
        {
            if (vidIsOpen)
                return;

            if (!libvlcIsOpen)
            {
                if (TryVlcOpen())
                {
                    libvlcIsOpen = true;
                    vlcDirTextBox.IsEnabled = false;
                    browseVlcButton.IsEnabled = false;
                }
                else
                {
                    MessageBox.Show("Error : invalid VLC install directory. \n\n" + 
                        "If you don't have VLC, please download it from http://videolan.org\n" +
                        "Also make sure that you point to a version of VLC with the same architecture " +
                        "as the current Buzz build (32 / 64 bit).",
                        "Videosync error");
                    return;
                }
            }

            // wave dead chicken around against latency
            string[] argv = { "--vout=" + (vlcOutCombo.Items[syncmac.VLCVout] as ComboBoxItem).Content.ToString(), "--clock-jitter=0", "--input-fast-seek", "--file-caching=10" };
            instance = Vlc.libvlc_new(4, argv);
            media = Vlc.libvlc_media_new_path(instance, syncmac.VideoPath);
            if (media == IntPtr.Zero)
            {
                MessageBox.Show("Error loading video.", "Videosync error");
                Vlc.libvlc_release(instance);
                return;
            }
            player = Vlc.libvlc_media_player_new_from_media(media);

            // embed vlc player into new owned window
            twin = new VideoWin(this);
            twin.Show();
            var wih = new WindowInteropHelper(twin);
            Vlc.libvlc_media_player_set_hwnd(player, wih.Handle);

            firstPlay = true;
            IntPtr evmgr = Vlc.libvlc_media_player_event_manager(player);
            Vlc.libvlc_event_attach(evmgr, (int)VlcEvent.MediaPlayerPlaying, vlcPlayingHandler, player);
            Vlc.libvlc_event_attach(evmgr, (int)VlcEvent.MediaPlayerEndReached, vlcEndHandler, player);
            Vlc.libvlc_event_attach(evmgr, (int)VlcEvent.MediaPlayerEncounteredError, vlcErrorHandler, player);
            Vlc.libvlc_media_player_play(player);
            vidIsOpen = true;
        }

        public void ReleaseVideo()
        {
            if (!vidIsOpen)
                return;
            Vlc.libvlc_media_player_stop(player);
            Vlc.libvlc_media_player_release(player);
            Vlc.libvlc_media_release(media);
            Vlc.libvlc_release(instance);
            vidIsOpen = false;
        }


        //      ============ GUI Bindings ============

        private void browseVideoButton_Click(object sender, RoutedEventArgs e)
        {
            var dlg = new Microsoft.Win32.OpenFileDialog();
            dlg.Title = "Choose Video File...";
            var result = dlg.ShowDialog();
            if (result == true)
                videoPathTextBox.Text = dlg.FileName;
        }

        private void browseVlcButton_Click(object sender, RoutedEventArgs e)
        {
            var dlg = new System.Windows.Forms.FolderBrowserDialog();
            dlg.Description = "Please specify the \"VLC\" install directory";
            var result = dlg.ShowDialog();
            if (result == System.Windows.Forms.DialogResult.OK)
            {
                vlcDirTextBox.Text = dlg.SelectedPath;
            }
        }

        private void videoPathTextBox_TextChanged(object sender, TextChangedEventArgs e)
        {
            syncmac.VideoPath = videoPathTextBox.Text;
        }

        private void vlcDirTextBox_TextChanged(object sender, TextChangedEventArgs e)
        {
            syncmac.VLCPath = vlcDirTextBox.Text;
            updatePrefs();
        }

        private void updatePrefs()
        {
            try
            {
                using (XmlWriter writer = XmlWriter.Create(Global.BuzzPath + "\\Gear\\Generators\\pac videosync config.xml"))
                {
                    writer.WriteStartElement("config");
                    writer.WriteElementString("vlcpath", syncmac.VLCPath);
                    writer.WriteElementString("vlcout", syncmac.VLCVout.ToString());
                    writer.WriteEndElement();
                    writer.Flush();
                    writer.Close();
                }
            }
            catch 
            {
                // vista permissions get out. can't find the exception type.
            }
        }

        private void manageBpmCheckBox_Changed(object sender, RoutedEventArgs e)
        {
            syncmac.ManageTempos = manageBpmCheckBox.IsChecked == true; // because nullable
            tempoDataGrid.IsEnabled = syncmac.ManageTempos;
        }

        private void volSlider_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            syncmac.VideoVolume = (int)volSlider.Value;
            volLabel.Content = String.Format("{0}%", syncmac.VideoVolume);
            if (vidIsOpen)
                Vlc.libvlc_audio_set_volume(player, (int)syncmac.VideoVolume);
        }

        private void openButton_Click(object sender, RoutedEventArgs e)
        {
            OpenVideo();
        }

        private void closeButton_Click(object sender, RoutedEventArgs e)
        {
            if (vidIsOpen)
                twin.Close();
        }

        private void vlcOutCombo_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            syncmac.VLCVout = vlcOutCombo.SelectedIndex;
            updatePrefs();
        }
    }

    public partial class VideoWin : Window
    {
        SyncGUI Parent;
        public IMachine Machine { get; set; }

        public VideoWin(SyncGUI parent)
        {
            Parent = parent;
            this.Owner = Window.GetWindow(Parent);
            this.Title = "Videosync";
            this.Width = 380;
            this.Height = 214;
            this.ShowInTaskbar = false;
            this.Closing += VideoWin_Closing;
        }

        private void VideoWin_Closing(object sender, CancelEventArgs e)
        {
            Parent.ReleaseVideo();
        }
    }
    
}
